#ifdef WIN32
#include <fstream>
#else
#include <fstream.h>
#endif
using namespace std;
//#include <fstream.h>

#include "imagetools.H"

//---------------------------------------------
//     IO_image::ReadRawType
//
// lecture d'une image au format raw sans entte,
// les donnes sont stockes au format du template
//---------------------------------------------
template <class T>
int ImageIO<T>::ReadRawType(istream &is, Image<T> &image, int largeur, int hauteur)
{
  image.Resize(hauteur,largeur);
  is.read((char *)(image.PixelPointer(0)), largeur*hauteur*sizeof(T));

  if (is.fail())
  {
    cerr << "ERREUR: impossible de lire une image dans un flux!" << endl;
    abort();
    return 0;
  }
  return 1;
}

template <class T>
int ImageIO<T>::ReadRawType(const char *const filename, Image<T> &image, int largeur, int hauteur)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return ReadRawType(ifs, image, largeur, hauteur);
}

//---------------------------------------------
//     ImageIO::ReadRawUChar
//
// lecture d'une image au format raw sans entte,
// les donnes sont stockes en unsigned char
//---------------------------------------------
template <class T>
int ImageIO<T>::ReadRawUChar(istream &is, Image<T> &image, int largeur, int hauteur)
{
  unsigned char *buffer;
  int i;

  buffer = new unsigned char [largeur*hauteur];
  is.read((char *)buffer, largeur*hauteur);

  if (is.fail())
  {
    cerr << "ERREUR: impossible de lire une image dans un flux!" << endl;
    abort();
    delete [] buffer;
    return 0;
  }

  image.Resize(hauteur,largeur);
  for(i=0; i<largeur*hauteur; i++)
    image[i] = (T) (buffer[i]);

  delete [] buffer;
  return 1;
}

template <class T>
int ImageIO<T>::ReadRawUChar(const char *const filename, Image<T> &image, int largeur, int hauteur)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return ReadRawUChar(ifs, image, largeur, hauteur);
}

//---------------------------------------------
//     ImageIO::ReadRawShort
//
// lecture d'une image au format raw sans entte,
// les donnes sont stockes en short
//---------------------------------------------
template <class T>
int ImageIO<T>::ReadRawShort(istream &is, Image<T> &image, int largeur, int hauteur)
{
  short *buffer;
  int i;

  buffer = new short [largeur*hauteur];
  is.read((char *)buffer, largeur*hauteur*sizeof(short));

  if (is.fail())
  {
    cerr << "ERREUR: impossible de lire une image dans un flux!" << endl;
    abort();
    delete [] buffer;
    return 0;
  }
  image.Resize(hauteur,largeur);
  for(i=0; i<largeur*hauteur; i++)
    image[i] = (T) (buffer[i]);

  delete [] buffer;
  return 1;
}

template <class T>
int ImageIO<T>::ReadRawShort(const char *const filename, Image<T> &image, int largeur, int hauteur)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return ReadRawShort(ifs, image, largeur, hauteur);
}



//---------------------------------------------
//     ImageIO<T>::SkipPGMWhiteSpace
//
// ignaration de tous les "blancs" dans une entte PGM
//  les commentaires sont aussi skipps
//  Une erreur est gnr si rien n'a t lu
//---------------------------------------------
template <class T>
void ImageIO<T>::SkipPGMWhiteSpace(istream &is)
{
  // nombre de caractres absorbs
  int skipped=0;
  int byteRead;

  for( byteRead = is.peek();
       (byteRead==' ')||(byteRead=='\n')||(byteRead==0)||(byteRead=='#')||(byteRead=='\t');
       byteRead = is.peek()
       )
  {
    skipped ++;
    if (byteRead=='#')
    {// absorbtion jusqu' la fin de la ligne
      is.ignore(100,'\n');
      // REMARQUE:
      //   il n'est pas ncessaire de mettre un taille dlirante, car une ligne
      //   de l'entte ne doit pas tes normalement suprieure  70 caractres.
    }
    else
      // skip du caractre
      is.get();
  }

  if (skipped==0)
  {
    cerr << "ERREUR: demande de lecture d'un whitespace dans l'entete, et rien n'a ete lu!!" << endl;
    abort();
  }
}

//---------------------------------------------
//     ImageIO::ReadEntetePGM
//
// lecture de l'entete d'une image au format PGM
//---------------------------------------------
template <class T>
int ImageIO<T>::ReadEntetePGM(istream &is, int &largeur, int &hauteur)
{
  int n;
  //  const int taille_buf=2000;

  // lecture du mot cl P5
  if ((is.get()!='P')||(is.get()!='5'))
  {
    cerr << "ERREUR : lecture de l'entete PGM, ne trouve pas P5!" << endl;
    abort();
  }

  // suppression des ventuels retour chariot ou espaces situ aprs P5
  //   if(ifs.get()!='\n')
  //   {
  //     cerr << "ERREUR : lecture de l'entete PGM, pas de retour chariot apres P5!" << endl;
  //     abort();
  //   }
  //   //lecture des commentaires
  //   int  comment=0;
  //   while(ifs.peek()=='#') {
  //     comment=1;
  //     ifs.ignore(taille_buf,'\n');
  //   }

  SkipPGMWhiteSpace(is);
  // lecture de la taille et du niveau de gris maximal
  is >> largeur;
  SkipPGMWhiteSpace(is);
  is >> hauteur;
  SkipPGMWhiteSpace(is);
  is >> n;

  if (is.fail())
  {
    cerr << "ERREUR: lecture de l'entete PGM, "
	 << "la lecture des dimensions et du niveau de gris maximum n'a pas pu etre effectuee correctement!" << endl;
    abort();
  }

  if(n>255)
  {
    cerr << "ERREUR : lecture de l'entete PGM, le niveau de gris maximal est superieur a 255!" << endl;
    abort();
  }


 // lis le retour chariot aprs 255 (c'est un retour chariot ncessairement, car c'est un PGM avec RAWBYTES - P5 type )
  if(is.get()!='\n')
  {
    cerr << "ERREUR : lecture de l'entete PGM, pas de retour chariot apres le niveau de gris maximal!" << endl;
    abort();
  }

  return 1;
}


//---------------------------------------------
//     ImageIO::ReadEntetePPM
//
// lecture de l'entete d'une image au format PPM
//---------------------------------------------
template <class T>
int ImageIO<T>::ReadEntetePPM(istream &is, int &largeur, int &hauteur)
{
  int n;
  //  const int taille_buf=2000;

  // lecture du mot cl P6
  if ((is.get()!='P')||(is.get()!='6'))
  {
    cerr << "ERREUR : lecture de l'entete PPM, ne trouve pas P6!" << endl;
    abort();
  }

  SkipPGMWhiteSpace(is);
  // lecture de la taille et du niveau de gris maximal
  is >> largeur;
  SkipPGMWhiteSpace(is);
  is >> hauteur;
  SkipPGMWhiteSpace(is);
  is >> n;

  if (is.fail())
  {
    cerr << "ERREUR: lecture de l'entete PPM, "
	 << "la lecture des dimensions et du niveau de gris maximum n'a pas pu etre effectuee correctement!" << endl;
    abort();
  }

  if(n>255)
  {
    cerr << "ERREUR : lecture de l'entete PPM, le niveau de gris maximal est superieur a 255!" << endl;
    abort();
  }


 // lis le retour chariot aprs 255 (c'est un retour chariot ncessairement, car c'est un PGM avec RAWBYTES - P5 type )
  if(is.get()!='\n')
  {
    cerr << "ERREUR : lecture de l'entete PPM, pas de retour chariot apres le niveau de gris maximal!" << endl;
    abort();
  }

  return 1;
}

//---------------------------------------------
//     ImageIO::InfoImagePGM
//
// lecture des infos d'une image au format avec entte PGM
//---------------------------------------------
template <class T>
int ImageIO<T>::InfoImagePGM(istream &is, int &largeur, int &hauteur)
{
  int continueReading;
  //-- lecture de l'entete
  continueReading = ReadEntetePGM(is, largeur, hauteur);

  return continueReading;
}

template <class T>
int ImageIO<T>::InfoImagePGM(const char *const filename, int &largeur, int &hauteur)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return InfoImagePGM(ifs, largeur, hauteur);

}


//---------------------------------------------
//     ImageIO::InfoImagePPM
//
// lecture des infos d'une image au format avec entte PPM
//---------------------------------------------
template <class T>
int ImageIO<T>::InfoImagePPM(istream &is, int &largeur, int &hauteur)
{
  int continueReading;
  //-- lecture de l'entete
  continueReading = ReadEntetePPM(is, largeur, hauteur);

  return continueReading;
}

template <class T>
int ImageIO<T>::InfoImagePPM(const char *const filename, int &largeur, int &hauteur)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return InfoImagePPM(ifs, largeur, hauteur);

}


//---------------------------------------------
//     ImageIO::ReadPGMType
//
// lecture d'une image au format avec entte PGM,
// les donnes sont stockes au format du template
//---------------------------------------------
template <class T>
int ImageIO<T>::ReadPGMType(istream &is, Image<T> &image)
{
  int continueReading;
  //-- lecture de l'entete
  int largeur,hauteur;
  continueReading = ReadEntetePGM(is, largeur, hauteur);

  //-- lecture des donnes
  if (continueReading)
    return ReadRawType(is, image, largeur, hauteur);
  else
    return 0;
}

template <class T>
int ImageIO<T>::ReadPGMType(const char *const filename, Image<T> &image)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return ReadPGMType(ifs, image);
}


//---------------------------------------------
//     ImageIO::ReadPGMUChar
//
// lecture d'une image au format avec entte PGM,
// les donnes sont stockes en unsigned char
//---------------------------------------------
template <class T>
int ImageIO<T>::ReadPGMUChar(istream &is, Image<T> &image)
{
  int continueReading;
  //-- lecture de l'entete
  int largeur,hauteur;
  continueReading = ReadEntetePGM(is, largeur, hauteur);

  //-- lecture des donnes
  if (continueReading)
    return ReadRawUChar(is, image, largeur, hauteur);
  else
    return 0;
}

template <class T>
int ImageIO<T>::ReadPGMUChar(const char *const filename, Image<T> &image)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return ReadPGMUChar(ifs, image);
}


//---------------------------------------------
//     ImageIO::ReadPGMShort
//
// lecture d'une image au format avec entte PGM,
// les donnes sont stockes en short
//---------------------------------------------
template <class T>
int ImageIO<T>::ReadPGMShort(istream &is, Image<T> &image)
{
  int continueReading;
  //-- lecture de l'entete
  int largeur,hauteur;
  continueReading = ReadEntetePGM(is, largeur, hauteur);

  //-- lecture des donnes
  if (continueReading)
    return ReadRawShort(is, image, largeur, hauteur);
  else
    return 0;
}

template <class T>
int ImageIO<T>::ReadPGMShort(const char *const filename, Image<T> &image)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return ReadPGMShort(ifs, image);
}


//---------------------------------------------
//     ImageIO::ReadRawRGBUChar
//
// lecture d'une image RGB au format raw sans entte,
// les donnes sont stockes en unsigned char
//---------------------------------------------
template <class T>
int ImageIO<T>::ReadRawRGBUChar(istream &is, Image<T> &imageR,Image<T> &imageG,Image<T> &imageB, int largeur, int hauteur)
{
  unsigned char *buffer;
  int i,j;

  buffer = new unsigned char [3*largeur*hauteur];
  is.read((char *)buffer, 3*largeur*hauteur);

  if (is.fail())
  {
    cerr << "ERREUR: impossible de lire une image dans un flux!" << endl;
    abort();
    delete [] buffer;
    return 0;
  }
  imageR.Resize(hauteur,largeur);
  imageG.Resize(hauteur,largeur);
  imageB.Resize(hauteur,largeur);
  for(i=0, j=0; i<largeur*hauteur; i++)
  {
    imageR[i] = (T) (buffer[j++]);
    imageG[i] = (T) (buffer[j++]);
    imageB[i] = (T) (buffer[j++]);
  }
  delete [] buffer;
  return 1;
}

template <class T>
int ImageIO<T>::ReadRawRGBUChar(const char *const filename, Image<T> &imageR,Image<T> &imageG,Image<T> &imageB, int largeur, int hauteur)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return ReadRawRGBUChar(ifs, imageR, imageG, imageB, largeur, hauteur);
}

//---------------------------------------------
//     ImageIO::ReadPPM
//
// lecture d'une image au format avec entte PPM,
//
//---------------------------------------------
template <class T>
int ImageIO<T>::ReadPPM(istream &is, Image<T> &imageR,Image<T> &imageG,Image<T> &imageB)
{
  int continueReading;
  //-- lecture de l'entete
  int largeur,hauteur;
  continueReading = ReadEntetePPM(is, largeur, hauteur);

  //-- lecture des donnes
  if (continueReading)
    return ReadRawRGBUChar(is, imageR, imageG, imageB, largeur, hauteur);
  else
    return 0;
}

template <class T>
int ImageIO<T>::ReadPPM(const char *const filename,  Image<T> &imageR,Image<T> &imageG,Image<T> &imageB)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return ReadPPM(ifs, imageR, imageG, imageB);
}

template <class T>
int ImageIO<T>::ReadPGM(const char *const filename, Image<T> &image)
{
  return ReadPGMUChar(filename, image);
}


template <class T>
int ImageIO<T>::WritePGM(const char *const filename, const Image<T> &image)
{
  return WritePGMUChar(filename, image);
}

template <>
int ImageIO<RGB>::ReadPPM(const char *const filename,  Image<RGB> &image)
{
  ifstream ifs(filename, ios::in|ios::binary);
  //ifstream ifs(filename);
  if (ifs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }
  int largeur,hauteur;
  int continueReading = ReadEntetePPM(ifs, largeur, hauteur);
  if (!continueReading)
    return 0;

  unsigned char *buffer;
  int i,j;

  buffer = new unsigned char [3*largeur*hauteur];
  ifs.read((char *)buffer, 3*largeur*hauteur);

  if (ifs.fail())
  {
    cerr << "ERREUR: impossible de lire une image dans un flux!" << endl;
    abort();
    delete [] buffer;
    return 0;
  }
  image.Resize(hauteur,largeur);
  for(i=0, j=0; i<largeur*hauteur; i++)
  {
    image[i].R = (octet) (buffer[j++]);
    image[i].G = (octet) (buffer[j++]);
    image[i].B = (octet) (buffer[j++]);
  }
  delete [] buffer;
  return 1;
}





template <class T>
int ImageIO<T>::WriteRawType(const char *const filename, const Image<T> &image)
{
  ofstream ofs(filename, ios::out|ios::binary);
  //ofstream ofs(filename);
  if (ofs.fail())
  {
    cerr << "ERREUR: impossible de creer le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return WriteRawType(ofs, image);
}

//---------------------------------------------
//     ImageIO::WriteRawUChar
//
// criture d'une image au format raw sans entte,
// les donnes sont stockes en unsigned char
//---------------------------------------------
template <class T>
int ImageIO<T>::WriteRawUChar(ostream &os, const Image<T> &image)
{
  unsigned char *buffer;
  int i;
  int largeur=image.NbCol();
  int hauteur=image.NbRow();
  buffer = new unsigned char [largeur*hauteur];
  for(i=0; i<largeur*hauteur; i++)
    //    if (image[i]<=0)
    //        buffer[i] = 0;
    //      else
    //        if (image[i]>=255)
    //  	buffer[i] = 255;
    //        else
    buffer[i] = (unsigned char) image[i];

  os.write((char *)buffer, largeur*hauteur);
  delete [] buffer;
  if (os.fail())
  {
    cerr << "ERREUR: impossible d'ecrire une image dans un flux!" << endl;
    abort();
    return 0;
  }
  return 1;
}

template <class T>
int ImageIO<T>::WriteRawUChar(const char *const filename, const Image<T> &image)
{
  ofstream ofs(filename, ios::out|ios::binary);
  //ofstream ofs(filename);
  if (ofs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return WriteRawUChar(ofs, image);
}

//---------------------------------------------
//     ImageIO::WriteRawShort
//
// criture d'une image au format raw sans entte,
// les donnes sont stockes en short
//---------------------------------------------
template <class T>
int ImageIO<T>::WriteRawShort(ostream &os, const Image<T> &image)
{
  short *buffer;
  int i;
  int largeur=image.NbCol();
  int hauteur=image.NbRow();
  buffer = new short [largeur*hauteur];
  for(i=0; i<largeur*hauteur; i++)
    buffer[i] = (short) (image[i]);
  os.write((char *)buffer, largeur*hauteur*sizeof(short));
  delete [] buffer;
  if (os.fail())
  {
    cerr << "ERREUR: impossible d'ecrire une image dans un flux!" << endl;
    abort();
    return 0;
  }
  return 1;
}

template <class T>
int ImageIO<T>::WriteRawShort(const char *const filename, const Image<T> &image)
{
  ofstream ofs(filename, ios::out|ios::binary);
  //ofstream ofs(filename);
  if (ofs.fail())
  {
    cerr << "ERREUR: impossible de creer le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return WriteRawShort(ofs, image);
}


//---------------------------------------------
//     ImageIO::WriteEntetePGM
//
// criture de l'entete d'une image au format PGM
//---------------------------------------------
template <class T>
int ImageIO<T>::WriteEntetePGM(ostream &os, const  int largeur,const int hauteur)
{
  // criture du mot cl P5
  os << "P5" << endl;

  //ajout des commentaires
  os << "# (c) ImageIO v1" << endl;

  // lecture de la taille et d'un troisime nombre
  os << largeur << " " << hauteur << endl << 255 << endl;

  return 1;
}

//---------------------------------------------
//     ImageIO::WriteEntetePseudoPGM
//
// criture de l'entete d'une image au format PGM
//---------------------------------------------
template <class T>
int ImageIO<T>::WriteEntetePseudoPGM(ostream &os, const int largeur, const int hauteur)
{
  // criture du mot cl P5
  os << "P5" << endl;

  //ajout des commentaires
  os << "# (c) ImageIO v1" << endl;
  if (sizeof(T)!=sizeof(unsigned char))
    os << "# ATTENTION: ceci n'est pas une image PGM classique, un pixel est stocke sur "
       << sizeof(T)*8 << " bits!" << endl;

  // lecture de la taille et d'un troisime nombre
  os << largeur << " " << hauteur << endl << 255 << endl;

  return 1;
}

//---------------------------------------------
//     ImageIO::WriteEntetePPM
//
// criture de l'entete d'une image au format PPM
//---------------------------------------------
template <class T>
int ImageIO<T>::WriteEntetePPM(ostream &os, const int largeur, const int hauteur)
{
  // criture du mot cl P5
  os << "P6" << endl;

  //ajout des commentaires
  os << "# (c) ImageIO v1" << endl;

  // lecture de la taille et d'un troisime nombre
  os << largeur << " " << hauteur << endl << 255 << endl;

  return 1;
}

//---------------------------------------------
//     ImageIO::WritePGMType
//
// criture d'une image au format avec entte PGM,
// les donnes sont stockes au format du template
//---------------------------------------------
template <class T>
int ImageIO<T>::WritePGMType(ostream &os, const Image<T> &image)
{
  //-- criture de l'entete
  WriteEntetePseudoPGM(os, image.NbCol(), image.NbRow());

  //-- criture des donnes
  return WriteRawType(os, image);
}

template <class T>
int ImageIO<T>::WritePGMType(const char *const filename, const Image<T> &image)
{
  ofstream ofs(filename, ios::out|ios::binary);
  //ofstream ofs(filename);
  if (ofs.fail())
  {
    cerr << "ERREUR: impossible de creer le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return WritePGMType(ofs, image);
}


//---------------------------------------------
//     ImageIO::WritePGMUChar
//
// criture d'une image au format avec entte PGM,
// les donnes sont stockes en unsigned char
//---------------------------------------------
template <class T>
int ImageIO<T>::WritePGMUChar(ostream &os, const Image<T> &image)
{
  //-- criture de l'entete
  WriteEntetePGM(os, image.NbCol(),image.NbRow());

  //-- criture des donnes
  return WriteRawUChar(os, image);
}

template <class T>
int ImageIO<T>::WritePGMUChar(const char *const filename, const Image<T> &image)
{
  ofstream ofs(filename, ios::out|ios::binary);
  //ofstream ofs(filename);
  if (ofs.fail())
  {
    cerr << "ERREUR: impossible de creer le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return WritePGMUChar(ofs, image);
}


//---------------------------------------------
//     ImageIO::WritePGMShort
//
// criture d'une image au format avec entte PGM,
// les donnes sont stockes en short
//---------------------------------------------
template <class T>
int ImageIO<T>::WritePGMShort(ostream &os, const Image<T> &image)
{
  //-- criture de l'entete
  WriteEntetePseudoPGM(os, image.NbCol(),image.NbRow());

  //-- criture des donnes
  return WriteRawShort(os, image);
}

template <class T>
int ImageIO<T>::WritePGMShort(const char *const filename, const Image<T> &image)
{
  ofstream ofs(filename, ios::out|ios::binary);
  //ofstream ofs(filename);
  if (ofs.fail())
  {
    cerr << "ERREUR: impossible de creer le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }

  return WritePGMShort(ofs, image);
}

//---------------------------------------------
//     ImageIO::WriteRawRGBUChar
//
// ecriture d'une image RGB au format raw sans entte,
// les donnes sont stockes en unsigned char
//---------------------------------------------
template <class T>
int ImageIO<T>::WriteRawRGBUChar(ostream &os, const Image<T> &imageR,const Image<T> &imageG,const Image<T> &imageB)
{
  unsigned char *buffer;
  int i,j;
  int largeur=imageR.NbCol();
  int hauteur=imageR.NbRow();

  buffer = new unsigned char [3*largeur*hauteur];
  for(i=0, j=0; i<largeur*hauteur; i++)
  {
    buffer[j++] = (unsigned char) imageR[i];
    buffer[j++] = (unsigned char) imageG[i];
    buffer[j++] = (unsigned char) imageB[i];
  }
  os.write((char *)buffer, 3*largeur*hauteur);
  delete [] buffer;
  if (os.fail())
  {
    cerr << "ERREUR: impossible d'ecrire une image dans un flux!" << endl;
    abort();
    return 0;
  }
  return 1;
}

template <class T>
int ImageIO<T>::WriteRawRGBUChar(const char *const filename, const Image<T> &imageR,const Image<T> &imageG,const Image<T> &imageB)
{
  ofstream ofs(filename, ios::out|ios::binary);
  //ifstream ifs(filename);
  if (ofs.fail())
  {
    cerr << "ERREUR: impossible d'ouvrir le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }
  return WriteRawRGBUChar(ofs, imageR, imageG, imageB);
}


//---------------------------------------------
//     ImageIO::WritePPM
//
// criture d'une image au format avec entte PPM,
// les donnes sont stockes en unsigned char
//---------------------------------------------
template <class T>
int ImageIO<T>::WritePPM(ostream &os, const Image<T> &imageR,const Image<T> &imageG,const Image<T> &imageB)
{
  int largeur=imageR.NbCol();
  int hauteur=imageR.NbRow();
  //-- criture de l'entete
  WriteEntetePPM(os, largeur,hauteur);
  //-- criture des donnes
  return WriteRawRGBUChar(os, imageR, imageG, imageB);
}

template <class T>
int ImageIO<T>::WritePPM(const char *const filename, const Image<T> &imageR,const Image<T> &imageG,const Image<T> &imageB)
{
  ofstream ofs(filename, ios::out|ios::binary);
  //ofstream ofs(filename);
  if (ofs.fail())
  {
    cerr << "ERREUR: impossible de creer le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }
  return WritePPM(ofs, imageR, imageG, imageB);
}

template <>
int ImageIO<RGB>::WritePPM(const char *const filename, const Image<RGB> &image)
{
  ofstream ofs(filename, ios::out|ios::binary);
  //ofstream ofs(filename);
  if (ofs.fail())
  {
    cerr << "ERREUR: impossible de creer le fichier " << filename << "!" << endl;
    abort();
    return 0;
  }
  int largeur=image.NbCol();
  int hauteur=image.NbRow();
  //-- criture de l'entete
  WriteEntetePPM(ofs, largeur,hauteur);
  unsigned char *buffer;
  int i,j;

  buffer = new unsigned char [3*largeur*hauteur];
  for(i=0, j=0; i<largeur*hauteur; i++)
  {
    buffer[j++] = (unsigned char) image[i].R;
    buffer[j++] = (unsigned char) image[i].G;
    buffer[j++] = (unsigned char) image[i].B;
  }

  ofs.write((char *)buffer, 3*largeur*hauteur);
  delete [] buffer;
  if (ofs.fail())
  {
    cerr << "ERREUR: impossible d'ecrire une image dans un flux!" << endl;
    abort();
    return 0;
  }
  return 1;
}


