#ifdef WIN32
#include <iostream>
#else
#include <iostream.h>
#endif


#ifdef PI
#  define MY_PI PI
#else
#  define MY_PI 3.14159265358979323846
#endif

using namespace std;

#include <math.h>
#include <map>
#include <vector>
#include <algorithm>

#define _ITCARRE(x) ((x)*(x))
#define _ITABS(x) ( (x)<0?-(x):(x) )
#define _ITMIN(a,b) ( (a)<(b)?(a):(b) )
#define _ITMAX(a,b) ( (a)>(b)?(a):(b) )

static double coef[8][3] = {
  {0.2125,0.7154,0.0721}, /* no sequence_display_extension */
  {0.2125,0.7154,0.0721}, /* ITU-R Rec. 709 (1990) */
  {0.299, 0.587, 0.114},  /* unspecified */
  {0.299, 0.587, 0.114},  /* reserved */
  {0.30,  0.59,  0.11},   /* FCC */
  {0.299, 0.587, 0.114},  /* ITU-R Rec. 624-4 System B, G */
  {0.299, 0.587, 0.114},  /* SMPTE 170M */
  {0.212, 0.701, 0.087}}; /* SMPTE 240M (1987) */

static int Inverse_Table_6_9[8][4] =
{
  {117504, 138453, 13954, 34903}, /* no sequence_display_extension */
  {117504, 138453, 13954, 34903}, /* ITU-R Rec. 709 (1990) */
  {104597, 132201, 25675, 53279}, /* unspecified */
  {104597, 132201, 25675, 53279}, /* reserved */
  {104448, 132798, 24759, 53109}, /* FCC */
  {104597, 132201, 25675, 53279}, /* ITU-R Rec. 624-4 System B, G */
  {104597, 132201, 25675, 53279}, /* SMPTE 170M */
  {117579, 136230, 16907, 35559}  /* SMPTE 240M (1987) */
};

template <class T>
T ImageTools<T>::bilinear(const Image<T> &image,const double yy,const double xx) {

  // test
  int nbc=image.NbCol();
  int nbl=image.NbRow();
  // pixel
  // 0 1
  // 2 3
  int i;
  int x[4],y[4];
  double dx,dy;
  x[0]=(int)(xx-0.5);
  y[0]=(int)(yy-0.5);
  dx=_ITABS(xx-(x[0]+0.5));
  dy=_ITABS(yy-(y[0]+0.5));
  x[1]=x[0]+1;
  y[1]=y[0];
  x[2]=x[0];
  y[2]=y[0]+1;
  x[3]=x[0]+1;
  y[3]=y[0]+1;
  for(i=0;i<4;i++) {
    if (x[i]<0) x[i]=0;
    if (y[i]<0) y[i]=0;
    if (x[i]>=nbc) x[i]=nbc-1;
    if (y[i]>=nbl) y[i]=nbl-1;
  }
  return((T)( ((1.0-dy)*(1.0-dx)) * image(y[0],x[0]) +
	      ((1.0-dy)*dx      ) * image(y[1],x[1]) +
	      ((1.0-dx)*dy      ) * image(y[2],x[2]) +
	      (dx*dy            ) * image(y[3],x[3]) )
	 );
}


template <class T>
T ImageTools<T>::getMax(const Image<T> &image)
{
#ifndef RELEASE
  assert(image.BufferSize()!=0);
#endif
  T maxim=image[0];
  for(int i=1;i<image.BufferSize();i++)
    maxim=(maxim<image[i]?image[i]:maxim);
  return(T(maxim));
}

template <class T>                                       // NB : ne marche qu avec octet !!!
double ImageTools<T>::getPSNR(const Image<T> &image1,const Image<T> &image2)
{
  double psnr=-1;
  if ((image1.BufferSize()!=0)&&
      (image1.NbCol()==image2.NbCol())&&
      (image1.NbRow()==image2.NbRow())
      )
  {
    double eqm=0.;
    unsigned int i,lg=image1.BufferSize();
    for(i=0;i<lg;i++)
      eqm+=(image1[i]-image2[i])*(image1[i]-image2[i]);
    psnr=10.0*log10(65025.0*lg/eqm);
  }
  return(psnr);
}

template <class T>
void ImageTools<T>::compacte(Image<T> &image) {

  if (image.BufferSize()!=0) {
    if ((int)getMax(image)>1023) {
      cerr<<" compacte : trop de regions"<<endl;
      return;
    }
  } else {
    cerr<<" compacte : imaeg vide"<<endl;
    return;
  }

  int tabInd[1024];
  int i,ind;
  int size=image.BufferSize();
  ind=0;
  for(i=0;i<1024;i++) tabInd[i]=-1;
  for(i=0;i<size;i++)
    if (tabInd[image[i]]==-1) {
      tabInd[image[i]]=ind;
      ind++;
    }
  for(i=0;i<size;i++)
    image[i]=tabInd[image[i]];
}

// dilate avec un _ITCARRE en 8 connexite
// l image doit etre en 0/1
template <class T>
void ImageTools<T>::dilate(Image<T> &in_image,const int taille) {

  if (in_image.BufferSize()!=0) {
    int nbc=in_image.NbCol();
    int nbl=in_image.NbRow();
    int i,j,u,v;
    int n2=taille/2;
    Image<T> source(nbl+n2*2,nbc+n2*2,(T)0);
    for(i=0;i<nbl;i++)
      for(j=0;j<nbc;j++)
	source(i+n2,j+n2)=in_image(i,j);

    for(i=0;i<nbl;i++)
      for(j=0;j<nbc;j++)
	for(u=0;u<taille;u++)
	  for(v=0;v<taille;v++)
	    if (source(i+u,j+v))
	      in_image(i,j)=(T)1;

  } else {
    cerr<<"dilate : pb in_image"<<endl;
  }
}

template <class T>
Image<octet> ImageTools<T>::functionHisto(const Image<T> &ima,const TypeConv8bits type) {

  Image<octet> sortie;

  if ((ima.BufferSize()==0)) {
    cerr<<"functionHisto : void image"<<endl;
    return(sortie);
  }
  int taille=ima.BufferSize();
  int h=ima.NbRow();
  int w=ima.NbCol();
  sortie.Resize(ima.NbRow(),ima.NbCol());
  vector<double> tab(h*w);
  map<double,int> lut;

  int i;
  for(i=0;i<h*w;i++)
    tab[i]=ima[i];
  sort(tab.begin(),tab.end());
  double pas;
  int k;
  double borne=0;
  double somme;

  switch(type) {
  case EGAL: // unif
    pas=(h*w)/256.;
    i=0;
    k=0;
    borne=pas;
    i=(int)borne-1;
    lut[tab[0]-1]=0;
    while(i<h*w&&k<256) {
      lut[tab[i]]=k;
      k++;
      borne+=pas;
      i=(int)borne-1;
    }
    lut[tab[h*w-1]+1]=255;
    break;

  case EXP:
    // petit precalcul
    somme=0.;
    for(i=0;i<256;i++)
      somme+=exp(i/255.);
    somme/=h*w;
    lut[tab[0]-1]=0;
    k=0;
    borne=exp(0/255.)/somme;
    i=(int)borne-1;

    while(i<h*w&&k<256) {
      lut[tab[i]]=k;
      k++;
      pas=exp(k/255.)/somme;
      borne+=pas;
      i=(int)borne-1;
    }
    lut[tab[h*w-1]+1]=255;
    break;

  case LOG:
    // petit precalcul
    somme=0.;
    for(i=0;i<256;i++)
      somme+=log((i/255.)+1.);
    somme/=h*w;
    lut[tab[0]-1]=0;
    k=0;
    borne=log((0/255.)+1.)/somme;
    i=(int)borne-1;
    while(i<h*w&&k<256) {
      lut[tab[i]]=k;
      k++;
      pas=log((k/255.)+1.)/somme;
      borne+=pas;
      i=(int)borne-1;
    }
    lut[tab[h*w-1]+1]=255;
    break;

  case EXPINV:
    // petit precalcul
    somme=0.;
    for(i=0;i<256;i++)
      somme+=exp(i/255.);
    somme/=h*w;
    lut[tab[0]-1]=0;
    k=0;
    borne=exp(255./255.)/somme;
    i=(int)borne-1;

    while(i<h*w&&k<256) {
      lut[tab[i]]=k;
      k++;
      pas=exp((255-k)/255.)/somme;
      borne+=pas;
      i=(int)borne-1;
    }
    lut[tab[h*w-1]+1]=255;
    break;

  case LOGINV:
    // petit precalcul
    somme=0.;
    for(i=0;i<256;i++)
      somme+=log((i/255.)+1.);
    somme/=h*w;
    lut[tab[0]-1]=0;
    k=0;
    borne=log(((255-0)/255.)+1.)/somme;
    i=(int)borne-1;
    while(i<h*w&&k<256) {
      lut[tab[i]]=k;
      k++;
      pas=log(((255-k)/255.)+1.)/somme;
      borne+=pas;
      i=(int)borne-1;
    }
    lut[tab[h*w-1]+1]=255;
    break;

  case SIN:
    // petit precalcul
    somme=0.;
    for(i=0;i<256;i++)
      somme+=sin((i/255.)*MY_PI);
    somme/=h*w;
    lut[tab[0]-1]=0;
    k=0;
    borne=sin((0/255.)*MY_PI)/somme;
    i=(int)borne-1;
    while(i<h*w&&k<256) {
      lut[tab[i]]=k;
      k++;
      pas=sin((k/255.)*MY_PI)/somme;
      borne+=pas;
      i=(int)borne-1;
    }
    lut[tab[h*w-1]+1]=255;
    break;

  case COS:
    // petit precalcul
    somme=0.;
    for(i=0;i<256;i++)
      somme+=_ITABS(cos((i/255.)*MY_PI));
    somme/=h*w;
    lut[tab[0]-1]=0;
    k=0;
    borne=_ITABS(cos((0/255.)*MY_PI))/somme;
    i=(int)borne-1;
    while(i<h*w&&k<256) {
      lut[tab[i]]=k;
      k++;
      pas=_ITABS(cos((k/255.)*MY_PI))/somme;
      borne+=pas;
      i=(int)borne-1;
    }
    lut[tab[h*w-1]+1]=255;
    break;

  default:
    cerr<<"functionHisto : unknown type "<<endl;
    break;
  }
  for(i=0;i<h*w;i++) {
    k=lut.lower_bound((double)ima[i])->second;
    sortie[i]=k;
  }
  return(sortie);
}

template <class T>
Image<octet> ImageTools<T>::convertTo8bits(const Image<T> &image1,const TypeConv8bits type,const double param) {

  Image<octet> sortie;

  if ((image1.BufferSize()==0)) {
    cerr<<"convertTo8bits : void image"<<endl;
    return(sortie);
  }
  int taille=image1.BufferSize();
  sortie.Resize(image1.NbRow(),image1.NbCol());
  T maxi;
  T mini;
  int i;
  double moy=0.0;
  double var=0.0;
  double val;
  maxi=image1[0];
  mini=image1[0];
  for(i=0;i<taille;i++) {
    if (image1[i]>maxi) maxi=image1[i];
    if (image1[i]<mini) mini=image1[i];
  }
  double ecart=(double)(maxi-mini);
  switch (type) {
  case NORMAL:
    for(i=0;i<taille;i++)
      sortie[i]=(octet) ( (double)(image1[i]-mini)*255.0/ecart );
    break;

  case SCALE:
    for(i=0;i<taille;i++) {
      val=( (double)(image1[i]-mini)*param );
      sortie[i]=(val<0.0)?0:((val>255.0)?255:(octet)val);
    }
    break;
  case LN1:
    for(i=0;i<taille;i++)
      sortie[i]=(octet)(367.9*log(1+( (double)(image1[i]-mini)/ecart )));
    break;

  case VAR1:
    for(i=0;i<taille;i++) {
      moy+=(double)image1[i];
      var+=_ITCARRE((double)image1[i]);
    }
    moy/=taille;
    var/=taille;
    var=sqrt(var-_ITCARRE(moy));

    for(i=0;i<taille;i++) {
      val=( ( (double)image1[i]-moy*128.0*param)/var)+128.0;
      //   cout<<"orig="<<image1(i)<<" val="<<val<<endl;
      sortie[i]=(val<0.0)?0:((val>255.0)?255:(octet)val);
    }

  case VAR2:
    for(i=0;i<taille;i++) {
      moy+=(double)image1[i];
      var+=_ITCARRE((double)image1[i]);
    }
    moy/=taille;
    var/=taille;
    var=sqrt(var-_ITCARRE(moy));
    for(i=0;i<taille;i++) {
      val=( (_ITABS((double)image1[i]-moy)*255.0*param)/var);
      //   cout<<"orig="<<image1(i)<<" val="<<val<<endl;
      sortie[i]=(val<0.0)?0:((val>255.0)?255:(octet)val);
    }
    break;

  default:
    cerr<<"convertTo8bits :  bad param"<<endl;
    break;
  }
  return(sortie);
}

template <>
void ImageTools<RGB>::RGBtoYUV(const Image<RGB> &ima, float *&imay,float *&imau,float *&imav,int &h,int &w,YUVSize size,YUVTransfo typetransfo) {

  double y, u, v;
  double cr, cg, cb, cu, cv;
  int i,j;
  cr = coef[typetransfo][0];
  cg = coef[typetransfo][1];
  cb = coef[typetransfo][2];
  cu = 0.5/(1.0-cb);
  cv = 0.5/(1.0-cr);
  h=ima.NbRow();
  w=ima.NbCol();
  int hu,wu;

  imay=new float[h*w];
  float *imaUt=new float[h*w];
  float *imaVt=new float[h*w];

  for(i=0;i<h*w;i++) {
    y = cr*ima[i].R + cg*ima[i].G + cb*ima[i].B;
    u = cu*(ima[i].B-y);
    v = cv*(ima[i].R-y);
    imay[i] = (219.0/256.0)*y + 16.5;  /* nominal range: 16..235 */
    imaUt[i] = (224.0/256.0)*u + 128.5; /* 16..240 */
    imaVt[i] = (224.0/256.0)*v + 128.5; /* 16..240 */
  }

  switch(size) {
  case S420:
    hu=h>>1;
    wu=w>>1;
    imau=new float[hu*wu];
    imav=new float[hu*wu];
    // decimation de U et V
    for(i=0;i<h-1;i++)
      for(j=0;j<w-1;j++) {
	imau[(i/2)*wu+(j/2)]=(((float)imaUt[i*w+j]+(float)imaUt[i*w+j+1]+(float)imaUt[(i+1)*w+j]+(float)imaUt[(i+1)*w+j+1]))/4.0;
	imav[(i/2)*wu+(j/2)]=(((float)imaVt[i*w+j]+(float)imaVt[i*w+j+1]+(float)imaVt[(i+1)*w+j]+(float)imaVt[(i+1)*w+j+1]))/4.0;
      }
    delete []imaUt;
    delete []imaVt;
    break;

  case S444:
    imau=imaUt;
    imav=imaVt;
    break;
  default:
    cerr<<"RGBtoYUV : pas fait !!"<<endl;
    return;
    break;
  }
}

template <>
Image<RGB> ImageTools<RGB>::YUVtoRGB(const float *const imay,const float *const imau,const float *const imav,const int h,const int w,YUVTransfo typetransfo,YUVSize size) {

  int i,j;
  int y, u, v, r, g, b;
  int crv, cbu, cgu, cgv;

  /* matrix coefficients */
  crv = Inverse_Table_6_9[typetransfo][0];
  cbu = Inverse_Table_6_9[typetransfo][1];
  cgu = Inverse_Table_6_9[typetransfo][2];
  cgv = Inverse_Table_6_9[typetransfo][3];

  float *imaUt=0;
  float *imaVt=0;
  int hu,wu;
  const float *py=imay;
  const float *pu;
  const float *pv;

  switch(size) {
  case S420:
    hu=h>>1;
    wu=w>>1;
    imaUt=new float[h*w];
    imaVt=new float[h*w];
    // interpolation de U et V => pour l instant simpliste
    for(i=0;i<h;i++)
      for(j=0;j<w;j++) {
	imaUt[i*w+j]=imau[(i/2)*wu+(j/2)];
	imaVt[i*w+j]=imav[(i/2)*wu+(j/2)];
      }
    pu=imaUt;
    pv=imaVt;
    break;

  case 444:
    pu=imau;
    pv=imav;
    break;
  default:
    cerr<<"RGBtoYUV : pas fait !! => 444"<<endl;
    pu=imau;
    pv=imav;
    break;
  }
  Image<RGB> imaout(h,w);

  for (i=0; i<h*w; i++,py++,pu++,pv++) {
    u = (int)(*pu) - 128;
    v = (int)(*pv) - 128;
    y = 76309 * ((int)(*py) - 16); /* (255/219)*65536 */
    // clip a la volee
    r = (y + crv*v + 32768)>>16;
    imaout[i].R=(unsigned char) ((r<0)?0:((r>255)?255:r));
    g= (y - cgu*u - cgv*v + 32768)>>16;
    imaout[i].G=(unsigned char) ((g<0)?0:((g>255)?255:g));
    b=(y + cbu*u + 32786)>>16;
    imaout[i].B=(unsigned char) ((b<0)?0:((b>255)?255:b));
    //   imaout(i).R = Clip[(y + crv*v + 32768)>>16];
    //      imaout(i).G = Clip[(y - cgu*u - cgv*v + 32768)>>16];
    //      imaout(i).B = Clip[(y + cbu*u + 32786)>>16];
  }

  switch(size) {
  case S420:
    delete []imaUt;
    delete []imaVt;
    break;

  case 444:
    break;
  default:
    cerr<<"RGBtoYUV : pas fait !! => 444"<<endl;
    break;
  }
  return(imaout);
}


#undef _ITCARRE
#undef _ITABS
#undef _ITMIN
#undef _ITMAX
