#ifndef _IMAGESTOOLS_MISC_H_
#define _IMAGESTOOLS_MISC_H_
#include <vector>
#include <algorithm>
#include <list>
#include <iterator>
#include <iostream>
#include <math.h>
using namespace std;

/** @name Classe d'outils divers
    @memo Classe d'outils divers (non encore classes)
     @doc
      Niveau de validite : moyen\\
      Niveau de performance :  moyen\\
      Niveau de finalisation : faible\\
    @author \URL[Bastien Peteuil]{mailto:bpeteuil@insa-rouen.fr}
    @author \URL[Franck Galpin]{mailto:fgalpinl@irisa.fr}
    \begin{verbatim}
    #include "imagetoolsmisc.H"
    \end{verbatim}
*/


//@{
/** Classe ImageToolsMisc */
template <class T>
class ImageToolsMisc
{
public:
  /** @doc Methode de sortie d'histogramme
      @param image une instance de la classe image
      @param nb_classes Nombre de classes dans le cas d'une image de flottants
      @param max Valeur maximum
      @param min Valeur mimimum
      @return un vecteur correspondant a l'histogramme
   */
  static vector<unsigned int> Histogramme(const Image<T> &image,const unsigned int nb_classes=256,const T min=0,const T max=255);

  /** @doc Methode de partitionnement en 4 connexit
      @param image une instance de la classe image
      @param nb_classes Le nombre de classes calcules
      @return l'image resultat avec un entier diffrent pour chaque partition
   */
  static Image<T> Partition_connexe(const Image<T> &image1,int &nb_classes); // en 4 connexite

};
//@}


template <class T>
vector<unsigned int> ImageToolsMisc<T>::Histogramme(const Image<T> &image,const unsigned int nb_classe,const T min,const T max)
{
  double pas;
  unsigned int k;
  int i,j;
  vector <unsigned int> histo(nb_classe+1);
  for(i=0;i<nb_classe;i++)
    histo[i]=0;

  pas=(double)(max-min)/(double)(nb_classe );
  for(i=0;i<image.NbCol();i++)
    {
      for(j=0;j<image.NbRow();j++)
	{
	  k=(unsigned int)((image(j,i) -min)/pas);

	  if(k>=nb_classe)
	    k=nb_classe-1;
	  if(k<0)
	    k=0;
	  histo[k]++;
	}
    }
  return histo;
}


template <class T>
Image<T> ImageToolsMisc<T>::Partition_connexe(const Image<T> &image1,int &nb_cl)
{
  int nbl=image1.NbRow();
  int nbc=image1.NbCol();
  Image<T> image(nbl,nbc);
  int courant=0;
  int centre=0;

  int i,j,k,u,l=1,i1,j1;
  bool test=false;

  image.InitTo(0);
  vector<list<int> > classes(256);
  vector<list<int> > classes_epure(256);
  list<int>::iterator iter,iter1;
  int  classe_cible;
  int nb_classes=0;

  /* initialisation des classes */
  for(i=0;i<nbl;i++)
  {
    for(j=0;j<nbc;j++)
    {

      if(courant>=250)
      {
	cout<<"plus de 250 classes"<<endl;
	l=1;
	nb_classes=0;
	for(u=1;u<=courant;u++)
	{
	  if(!classes[u].empty())
	  {
	    classes_epure[l].clear();
	    classes_epure[l]=classes[u];
	    classes[u].clear();
	    l++;
	    nb_classes++;
	  }
	}

	cout<<"nombre de classes apres epuration"<<nb_classes<<endl;
	for(u=1;u<=nb_classes;u++)
	{
	  classes[u]=classes_epure[u];
	  for(i1=0;i1<nbl;i1++)
	  {
	    for(j1=0;j1<nbc;j1++)
	    {
	      if(image(i1,j1)!=0)
	      {
		for(k=1;k<=nb_classes;k++)
		{
		  iter=find(classes_epure[k].begin(),classes_epure[k].end(),image(i1,j1));
		  if(iter!=classes_epure[k].end())
		  {
		    image(i1,j1)=k;
		  }
		}
	      }
	    }
	  }
	}
	  courant=nb_cl-1;
      }/* fin du cas explosion du nombre d'etiquettes */






      test=false;
      centre=0;
      classe_cible=0;

      if(image1(i,j)!=0)
      {
	if(i==0&&j==0)
	{
	  courant++;
	  classes[courant].push_back(courant);
	  image(i,j)=courant;
	}

	if(i==0&&j!=0)
	{
	  if(image(i,j-1)!=0)
	    image(i,j)=image(i,j-1);

	  else
	  {
	    courant++;
	    classes[courant].push_back(courant);
	    image(i,j)=courant;
	  }
	}


	if(j==0&&i!=0)
	{
	  if(image(i-1,j)!=0)
	    image(i,j)=image(i-1,j);

	  else
	  {
	    courant++;
	    classes[courant].push_back(courant);
	    image(i,j)=courant;
	  }
	}




	if(i!=0&&j!=0)
	{

	  if((image(i-1,j)!=0)&&(image(i,j-1)==0))
	  {
	    test=true;
	    centre=image(i-1,j);
	  }

	  if((image(i-1,j)==0)&&(image(i,j-1)!=0))
	  {
	    test=true;
	    centre=image(i,j-1);
	  }

	  /* on a une region de contact entre deux classes */
	  if((image(i,j-1)!=0)&&(image(i-1,j)!=0))
	  {
	    test=true;
	    centre=image(i-1,j);
	    for(k=1;k<=courant;k++)
	    {
	      iter=find(classes[k].begin(),classes[k].end(),image(i-1,j));
	      if(iter!=classes[k].end())
	      {
		classe_cible=k;
		break;
	      }
	    }
	    /* on ajoute l'element de la classe */
	    for(k=1;k<=courant;k++)
	    {
	      iter=find(classes[k].begin(),classes[k].end(),image(i,j-1));

	      if(iter!=classes[k].end()&&(k!=classe_cible))
	      {
		iter1=classes[k].begin();

		while(iter1!=classes[k].end())
		{

		  classes[classe_cible].push_back(*iter1);
		  iter1++;
		}
		classes[k].clear();
		break;
	      }
	    }


	  }

	  /* si on n'a pas trouver d'elements marques dans la fenetre */
	  if(!test)
	  {
	    courant++;
	    classes[courant].push_back(courant);
	    image(i,j)=courant;
	  }

	  if(test)
	  {
	    image(i,j)=centre;
	  }

	}/* fin du if au milieu*/


      } /* fin du si element non vide */

    }
  }/*fin de boucle image */

  nb_classes=0;
  l=1;
  for(i=1;i<=courant;i++)
  {
    if(!classes[i].empty())
    {
      classes_epure[l].clear();
      classes_epure[l]=classes[i];
      l++;
      nb_classes++;
    }
  }

  // cout<<"Partition connexe : nombre de classes = "<<nb_classes<<endl;

  for(i=0;i<nbl;i++)
  {
    for(j=0;j<nbc;j++)
    {
      if(image(i,j)!=0)
      {
	for(k=1;k<=nb_classes;k++)
	{
	  iter=find(classes_epure[k].begin(),classes_epure[k].end(),image(i,j));
	  if(iter!=classes_epure[k].end())
	  {
	    image(i,j)=k;
	  }
	}
      }
    }
  }
  nb_cl=nb_classes;
  return image;
}


#endif
